# Copyright (c) 2023, BLUEWALM. All rights reserved. 
# Base embedding 

import torch
# 
import pytorch_embeddings.globals as globals
from pytorch_embeddings.NativeEmbedding import NativeEmbedding
from pytorch_embeddings.FrobeniusEmbedding import FrobeniusEmbedding
from pytorch_embeddings.initializers import frobenius_initializers_


class BaseEmbedding(torch.nn.Module):
    r""" BaseEmbedding is a wrapper class for all the embeddings in this library. 
       Internal. Only our context managers should use it. """
    def asserts(self):
        assert 0 < self.num_embeddings
        assert 0 < self.embedding_dim
    
    def __init__(self, num_embeddings, embedding_dim, **kwargs):
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.embedding = None
        self.sparse = kwargs.get('sparse', False)
        device = kwargs.get('device', 'cuda')
        dtype = kwargs.get('dtype', torch.float32)
        kwargs = {'device' : device, 'dtype' : dtype}
        indicator = torch.ones(1, **kwargs)
        indicator = torch.nn.Parameter(indicator)
        self.register_buffer('indicator', indicator)
        self.asserts()
        globals.callback_embeddings.append(self)
    
    def reset(self, embedding_type, **kwargs):
        kwargs.update({'device' : self.indicator.device, 'dtype' : self.indicator.dtype, 'sparse' : self.sparse})
        self.embedding = embedding_type(self.num_embeddings, self.embedding_dim, **kwargs)
    
    def __repr__(self):
        return "Embedding(" + self.extra_repr() + ")"
    
    def extra_repr(self):
        if self.embedding is None:
            s = 'embedding_type=uninitialized, num_embeddings={num_embeddings}, embedding_dim={embedding_dim}'
            return s.format(**self.__dict__)
        else:
            return self.embedding.extra_repr()
    
    def forward(self, index):
        return self.embedding(index)
    
    @property
    def weight(self):
        if isinstance(self.embedding, NativeEmbedding):
            return self.embedding.weight
        else:
            return self
    
    @weight.setter
    def weight(self, other):
        self.set_state_from_weight(other)
    
    @property
    def data(self):
        if isinstance(self.embedding, NativeEmbedding):
            return self.embedding.weight.data
        else:
            return self
    
    @data.setter
    def data(self, other):
        self.set_state_from_weight(other)
    
    def set_state_from_weight(self, other):
        if isinstance(self.embedding, NativeEmbedding):
            assert other.shape == (self.num_embeddings, self.embedding_dim)
            self.embedding.weight.data.copy_(other)
        else:
            raise Exception("setting the compressed embedding weight from an uncompressed weight is not possible. "\
                          "please try to pre-train the compressed embedding with the desired weight set as target. "\
                          "alternatively, try using one of the in-place random initializers. ")
    
    def set_state_from_random_source(self, original_function, function_name, args, kwargs):
        if isinstance(self.embedding, NativeEmbedding):
            original_function(self.embedding.weight.data, *args, **kwargs)
        elif isinstance(self.embedding, FrobeniusEmbedding):
            initializer_ = frobenius_initializers_[function_name]
            initializer_(self.embedding, *args, **kwargs)
        elif self.embedding is None:
            raise Exception("uninitialized embedding - need to specify the types and "\
                         "corresponding hyperparameters of all the embeddings")
        else:
            raise Exception("unknown embedding type")
    
    def uniform_(self, *args, **kwargs):
        def original_function(tensor, *args, **kwargs):
            return tensor.uniform_(*args, **kwargs)
        arguments = (original_function, 'uniform_', args, kwargs)
        initializer = (self.set_state_from_random_source, arguments)
        globals.callback_initializers.append(initializer)
    
    def normal_(self, *args, **kwargs):
        def original_function(tensor, *args, **kwargs):
            return tensor.normal_(*args, **kwargs)
        arguments = (original_function, 'normal_', args, kwargs)
        initializer = (self.set_state_from_random_source, arguments)
        globals.callback_initializers.append(initializer)

