# Copyright (c) 2023, BLUEWALM. All rights reserved. 
# Frobenius embedding 

import os
import torch
import pytorch_embeddings_extension
from .heuristics_frobenius import default_frobenius


class FrobeniusEmbedding(pytorch_embeddings_extension.FrobeniusEmbedding):
    r""" Frobenius Embedding that stores embeddings of a fixed dictionary and size. 
    
    The input to the module is a list of indices, and the output is the corresponding embeddings. 
    By default, the module is uninitialized. 
    
    Args:
        num_embeddings (int): size of the dictionary of embeddings. 
        embedding_dim (int): the size of each embedding vector. 
    kwArgs:
        frobenius_rank (int): a hyperparameter. correlates positively with the expressive capacity 
                          of the layer and its size. 
        frobenius_blocks (int): a hyperparameter. must be a divisor of embedding_dim. 
                            correlates positively with the expressive capacity of the layer and its size. 
        alternating_gradients (bool): setting it to `True` will result in a slower, 
                                but a more stable convergence. 
        onchip_memory (bool): setting it to `True` will result in a small latency drop, 
                            but it will not use any global memory. default: `True`
        silent (bool): setting it to `True` will result in the details of the embedding hidden. 
                    hyperparameters, size, etc. will not be printed to console when `print(model)` 
                    is called, respectively. default: `False`
        device (str): sets the device the embedding will be constructed on. default: `cuda`
        dtype (torch.dtype): sets the datatype of the embedding at construction time. default: `torch.float32`
    
    Shape:
        - Input: :math:`(*)`, LongTensor of arbitrary shape containing the indices to extract
        - Output: :math:`(*, H)`, where `*` is the input shape and :math:`H=\text{embedding\_dim}`
    
    Examples::
    
        >>> # a FrobeniusEmbedding module containing 128 vectors of size 4
        >>> embedding = FrobeniusEmbedding(128, 4, frobenius_rank=8, frobenius_blocks=2, device='cuda')
        >>> # a batch of 2 samples of 4 indices each
        >>> input = torch.LongTensor([[1, 2, 4, 5], [4, 3, 2, 9]])
        >>> embedding(input)
        tensor([[[-0.0251, -1.6902,  0.7172,  0.3434],
              [-0.6431,  0.0748,  0.6969,  0.1444],
              [ 1.4970,  1.3448, -0.9685,  0.1433],
              [-0.3677, -2.7265, -0.1685, -0.1434]],

              [[ 1.4970,  1.3448, -0.9685, -0.3434],
              [ 0.4362, -0.4004,  0.9400,  0.2734],
              [-0.6431,  0.0748,  0.6969,  0.8368],
              [ 0.9124, -2.3616,  1.1151, -0.8411]]])
    """
    def __init__(self, num_embeddings, embedding_dim, **kwargs):
        device = torch.device(kwargs.get('device', 'cuda'))
        dtype = kwargs.get('dtype', torch.float32)
        default = default_frobenius(num_embeddings, embedding_dim, dtype)
        frobenius_rank = kwargs.get('frobenius_rank', default.frobenius_rank)
        frobenius_blocks = kwargs.get('frobenius_blocks', default.frobenius_blocks)
        kwargs.update({'device' : device, 'dtype' : dtype, 
                     'frobenius_rank' : frobenius_rank, 
                     'frobenius_blocks' : frobenius_blocks})
        super().__init__(num_embeddings, embedding_dim, **kwargs)
    
    def __repr__(self):
        return "Embedding(" + self.extra_repr() + ")"
    
    def __str__(self):
        return "Embedding(" + self.extra_repr() + ")"

