# Copyright (c) 2023, BLUEWALM. All rights reserved. 
# Native embedding 

import torch


class NativeEmbedding(torch.nn.Embedding):
    r""" NativeEmbedding that stores embeddings of a fixed dictionary and size. 
        This class is a wrapper for `torch.nn.Embedding`. The difference is that this class 
        prints the size of the embedding when `print(model)` is called, respectively. 
        The input to the module is a list of indices, and the output is the corresponding embeddings. 
        By default, the module is initialized with the standard normal distribution. 
    
    Args:
        num_embeddings (int): size of the dictionary of embeddings. 
        embedding_dim (int): the size of each embedding vector. 
    kwArgs:
        sparse (bool): setting it to True means sparse embeddings will be used. default: `False`
        device (str): sets the device the embedding will be constructed on. default: `cuda`
        dtype (torch.dtype): sets the datatype of the embedding at construction time. default: `torch.float32`
    
    Shape:
        - Input: :math:`(*)`, LongTensor of arbitrary shape containing the indices to extract
        - Output: :math:`(*, H)`, where `*` is the input shape and :math:`H=\text{embedding\_dim}`
    
    Examples::
    
        >>> # a NativeEmbedding module containing 128 vectors of size 3
        >>> embedding = NativeEmbedding(128, 3, device='cpu', sparse=True)
        >>> # a batch of 2 samples of 4 indices each
        >>> input = torch.LongTensor([[1, 2, 4, 5], [4, 3, 2, 9]])
        >>> embedding(input)
        tensor([[[-0.0251, -1.6902,  0.7172],
              [-0.6431,  0.0748,  0.6969],
              [ 1.4970,  1.3448, -0.9685],
              [-0.3677, -2.7265, -0.1685]],

              [[ 1.4970,  1.3448, -0.9685],
              [ 0.4362, -0.4004,  0.9400],
              [-0.6431,  0.0748,  0.6969],
              [ 0.9124, -2.3616,  1.1151]]])
    """
    def __init__(self, num_embeddings, embedding_dim, **kwargs):
        device = kwargs.get('device', 'cuda')
        dtype = kwargs.get('dtype', torch.float32)
        sparse = kwargs.get('sparse', False)
        kwargs = {'device' : device, 'dtype' : dtype, 'sparse' : sparse}
        super().__init__(num_embeddings, embedding_dim, **kwargs)
    
    def __repr__(self):
        return "Embedding(" + self.extra_repr() + ")"
    
    def extra_repr(self):
        s = 'embedding_type=native, num_embeddings={num_embeddings}, embedding_dim={embedding_dim}'
        precision = {torch.float32 : 'fp32', torch.float16 : 'fp16'}
        precision = precision[self.weight.data.dtype]
        s += ', precision={precision}'
        size = self.weight.data.nelement() * self.weight.data.element_size()
        size = round(size / 1024**2, 4)
        s += ', size=' + str(size) + ' MB'
        return s.format(**self.__dict__, precision=precision)

