# Copyright (c) 2023, BLUEWALM. All rights reserved. 
# frobenius heuristics

import torch
from math import sqrt, log
from collections import namedtuple


# the list of embedding dims supported by the kernels
embedding_dims_supported = [8, 16, 32, 64, 128, 256]


def generate_configs_frobenius(num_embeddings, embedding_dim, precision):
    # define the method for filtering good frobenius_block values
    def good_frobenius_blocks(frobenius_blocks):
        condition1 = embedding_dim % frobenius_blocks == 0
        condition2 = (embedding_dim // frobenius_blocks) in embedding_dims_supported
        conditions = [condition1, condition2]
        return all(conditions)
    # compute possible values for frobenius_blocks 
    possible_frobenius_blocks = []
    for frobenius_blocks in range(1, embedding_dim+1):
        if good_frobenius_blocks(frobenius_blocks):
            possible_frobenius_blocks.append(frobenius_blocks)
    # compute FrobeniusEmbedding constants 
    n = int(round(sqrt(num_embeddings)))
    n = n + ((8 - n) % 8)
    # compute the size of the configs 
    def cfg_size(frobenius_rank, frobenius_blocks):
        size = n * frobenius_rank
        size += n * frobenius_rank
        size += frobenius_rank * (embedding_dim // frobenius_blocks)
        size = size * frobenius_blocks
        if precision in ('fp16', torch.float16):
            # everything is fp16 
            size = size * 2
        elif precision in ('fp32', torch.float32):
            # everything is fp32 
            size = size * 4
        return size
    # enumerate all the configurations we are interested in 
    configs = []
    _config = namedtuple('config', 'embedding_type num_embeddings embedding_dim size precision frobenius_rank frobenius_blocks')
    for frobenius_blocks in possible_frobenius_blocks:
        frobenius_ranks = []
        for b in range(2, 11):
            r = log(n * embedding_dim // frobenius_blocks, b)
            r = int(round(r))
            r_mod8 = r + ((8 - r) % 8)
            r_pow2 = 2**(int(log(r_mod8 - 1, 2)) + 1)
            frobenius_ranks.append(r_mod8)
            frobenius_ranks.append(r_pow2)
        frobenius_ranks = list(set(frobenius_ranks))
        for frobenius_rank in frobenius_ranks:
            size = cfg_size(frobenius_rank, frobenius_blocks)
            config = _config('frobenius', num_embeddings, embedding_dim, size, precision, frobenius_rank, frobenius_blocks)
            configs.append(config)
    return configs


def default_frobenius(num_embeddings, embedding_dim, precision):
    configs = generate_configs_frobenius(num_embeddings, embedding_dim, precision)
    configs.sort(key=lambda config : ((config.embedding_dim // config.frobenius_blocks) - 32)**2 + config.frobenius_rank)
    return configs[0]

