# Copyright (c) 2023, BLUEWALM. All rights reserved. 
# initializer wrappers

import math
import torch
from torch.nn.init import calculate_gain


# unsupported initializers


def eye_(*args, **kwargs):
    raise Exception("eye_  --  unsupported initialization method")


def dirac_(*args, **kwargs):
    raise Exception("dirac_  --  unsupported initialization method")


def orthogonal_(*args, **kwargs):
    raise Exception("orthogonal_  --  unsupported initialization method")


# supported initializers


def _calculate_fan_in_and_fan_out(num_embeddings, embedding_dim):
    fan_in = embedding_dim
    fan_out = num_embeddings
    return fan_in, fan_out


def _calculate_correct_fan(num_embeddings, embedding_dim, mode):
    mode = mode.lower()
    valid_modes = ['fan_in', 'fan_out']
    if mode not in valid_modes:
        raise ValueError("Mode {} not supported, please use one of {}".format(mode, valid_modes))
    
    fan_in, fan_out = _calculate_fan_in_and_fan_out(num_embeddings, embedding_dim)
    return fan_in if mode == 'fan_in' else fan_out


# FrobeniusEmbedding initializers


def frobenius_uniform_(embedding, a=0, b=1, eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values drawn from the uniform
    distribution :math:`\mathcal{U}(a, b)`.
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        a: the lower bound of the uniform distribution
        b: the upper bound of the uniform distribution
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_uniform_(w)
    """
    embedding.uniform_(a, b, eps=eps)


def frobenius_normal_(embedding, mean=0, std=1, eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values drawn from the normal
    distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`.
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution

    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_normal_(w)
    """
    embedding.normal_(mean, std, eps=eps)


def frobenius_trunc_normal_(embedding, mean=0, std=1, a=-2, b=2, eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values drawn from a truncated 
    normal distribution. The values are effectively drawn from the 
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)` 
    with values outside :math:`[a, b]` redrawn until they are within 
    the bounds. The method used for generating the random values works 
    best when :math:`a \leq \text{mean} \leq b`. 
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> frobenius_trunc_normal_(w)
    """
    embedding.trunc_normal_(mean, std, a, b, eps=eps)


def frobenius_constant_(embedding, val):
    r"""Initializes the input `FrobeniusEmbedding` with the value :math:`\text{val}`.
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        val: the value to fill the tensor with
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_constant_(w, 0.3)
    """
    embedding.constant_(val)


def frobenius_ones_(embedding):
    r"""Initializes the input `FrobeniusEmbedding` with the scalar value `1`.

    Args:
        embedding: an instance of a `FrobeniusEmbedding`
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_ones_(w)
    """
    embedding.constant_(1.0)


def frobenius_zeros_(embedding):
    r"""Initializes the input `FrobeniusEmbedding` with the scalar value `0`.

    Args:
        embedding: an instance of a `FrobeniusEmbedding`
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_zeros_(w)
    """
    embedding.constant_(0.0)


def frobenius_xavier_uniform_(embedding, gain=1, eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values according to the method
    described in `Understanding the difficulty of training deep feedforward
    neural networks` - Glorot, X. & Bengio, Y. (2010), using a uniform
    distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{U}(-a, a)` where
    
    .. math::
        a = \text{gain} \times \sqrt{\frac{6}{\text{fan\_in} + \text{fan\_out}}}
    
    Also known as Glorot initialization. 
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        gain: an optional scaling factor
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_xavier_uniform_(w, gain=nn.init.calculate_gain('relu'))
    """
    shape = (embedding.num_embeddings, embedding.embedding_dim)
    fan_in, fan_out = _calculate_fan_in_and_fan_out(*shape)
    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
    a = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation
    embedding.uniform_(-a, a, eps=eps)


def frobenius_xavier_normal_(embedding, gain=1, eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values according to the method
    described in `Understanding the difficulty of training deep feedforward
    neural networks` - Glorot, X. & Bengio, Y. (2010), using a normal
    distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{N}(0, \text{std}^2)` where

    .. math::
        \text{std} = \text{gain} \times \sqrt{\frac{2}{\text{fan\_in} + \text{fan\_out}}}

    Also known as Glorot initialization.

    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        gain: an optional scaling factor

    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_xavier_normal_(w)
    """
    shape = (embedding.num_embeddings, embedding.embedding_dim)
    fan_in, fan_out = _calculate_fan_in_and_fan_out(*shape)
    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
    embedding.normal_(0, std, eps=eps)


def frobenius_kaiming_uniform_(embedding, a=0, mode='fan_in', nonlinearity='leaky_relu', eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values according to the method
    described in `Delving deep into rectifiers: Surpassing human-level
    performance on ImageNet classification` - He, K. et al. (2015), using a
    uniform distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{U}(-\text{bound}, \text{bound})` where

    .. math::
        \text{bound} = \text{gain} \times \sqrt{\frac{3}{\text{fan\_mode}}}

    Also known as He initialization.

    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        a: the negative slope of the rectifier used after this layer (only
            used with ``'leaky_relu'``)
        mode: either ``'fan_in'`` (default) or ``'fan_out'``. Choosing ``'fan_in'``
            preserves the magnitude of the variance of the weights in the
            forward pass. Choosing ``'fan_out'`` preserves the magnitudes in the
            backwards pass.
        nonlinearity: the non-linear function (`nn.functional` name),
            recommended to use only with ``'relu'`` or ``'leaky_relu'`` (default).

    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_kaiming_uniform_(w, mode='fan_in', nonlinearity='relu')
    """
    shape = (embedding.num_embeddings, embedding.embedding_dim)
    fan = _calculate_correct_fan(*shape, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    bound = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation
    embedding.uniform_(-bound, bound, eps=eps)


def frobenius_kaiming_normal_(embedding, a=0, mode='fan_in', nonlinearity='leaky_relu', eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` with values according to the method
    described in `Delving deep into rectifiers: Surpassing human-level
    performance on ImageNet classification` - He, K. et al. (2015), using a
    normal distribution. The resulting tensor will have values sampled from
    :math:`\mathcal{N}(0, \text{std}^2)` where
    
    .. math::
        \text{std} = \frac{\text{gain}}{\sqrt{\text{fan\_mode}}}
    
    Also known as He initialization.
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        a: the negative slope of the rectifier used after this layer (only
            used with ``'leaky_relu'``)
        mode: either ``'fan_in'`` (default) or ``'fan_out'``. Choosing ``'fan_in'``
            preserves the magnitude of the variance of the weights in the
            forward pass. Choosing ``'fan_out'`` preserves the magnitudes in the
            backwards pass.
        nonlinearity: the non-linear function (`nn.functional` name),
            recommended to use only with ``'relu'`` or ``'leaky_relu'`` (default).
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_kaiming_normal_(w, mode='fan_out', nonlinearity='relu')
    """
    shape = (embedding.num_embeddings, embedding.embedding_dim)
    fan = _calculate_correct_fan(*shape, mode)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    embedding.normal_(0, std, eps=eps)


def frobenius_sparse_(embedding, sparsity, std=0.01, eps=1e-4):
    r"""Initializes the input `FrobeniusEmbedding` in a sparse way, where the
    non-zero elements will be drawn from the normal distribution
    :math:`\mathcal{N}(0, 0.01)`, as described in `Deep learning via
    Hessian-free optimization` - Martens, J. (2010).
    
    Args:
        embedding: an instance of a `FrobeniusEmbedding`
        sparsity: The fraction of elements in each column to be set to zero
        std: the standard deviation of the normal distribution used to generate
            the non-zero values
    
    Examples:
        >>> w = FrobeniusEmbedding(128, 32)
        >>> quotient_sparse_(w, sparsity=0.1)
    """
    embedding.sparse_(sparsity, std, eps=eps)


frobenius_initializers_ = {
            'uniform_'         : frobenius_uniform_, 
            'normal_'          : frobenius_normal_, 
            'trunc_normal_'     : frobenius_trunc_normal_, 
            'constant_'         : frobenius_constant_, 
            'ones_'            : frobenius_ones_, 
            'zeros_'           : frobenius_zeros_, 
            'xavier_uniform_'   : frobenius_xavier_uniform_, 
            'xavier_normal_'    : frobenius_xavier_normal_, 
            'kaiming_uniform_'  : frobenius_kaiming_uniform_, 
            'kaiming_normal_'   : frobenius_kaiming_normal_, 
            'sparse_'           : frobenius_sparse_, 
            # unsupported
            'eye_'            : eye_, 
            'dirac_'           : dirac_, 
            'orthogonal_'      : orthogonal_
}

